@@ -4,13 +4,16 @@
 #
 # Copyright (C) 2017, James R. Barlow (https://github.com/jbarlow83/)
 
+import base64
 import struct
 from abc import ABC, abstractmethod
+from collections import deque
 from decimal import Decimal
 from io import BytesIO
 from itertools import zip_longest
 from pathlib import Path
 from shutil import copyfileobj
+from typing import NamedTuple, Optional
 from zlib import decompress
 
 from PIL import Image, ImageCms
@@ -41,17 +44,21 @@ class NotExtractableError(Exception):
     "Indicates that an image cannot be directly extracted."
 
 
+class HifiPrintImageNotTranscodableError(NotExtractableError):
+    "Image contains high fidelity printing information and cannot be extracted."
+
+
 class InvalidPdfImageError(Exception):
     "This image is not valid according to the PDF 1.7 specification."
 
 
 def array_str(value):
-    """Simplify pikepdf objects to array of str. Keep Streams intact."""
+    """Simplify pikepdf objects to array of str. Keep Streams and dictionaries intact."""
 
     def _array_str(item):
         if isinstance(item, (list, Array)):
             return [_array_str(subitem) for subitem in item]
-        elif isinstance(item, (Stream, bytes, int)):
+        elif isinstance(item, (Stream, Dictionary, bytes, int)):
             return item
         elif isinstance(item, (Name, str)):
             return str(item)
@@ -86,14 +93,28 @@ def metadata_from_obj(obj, name, type_, default):
     raise NotImplementedError('Metadata access for ' + name)
 
 
+class PaletteData(NamedTuple):
+    """Returns the color space and binary representation of the palette.
+
+    ``base_colorspace`` is typically ``"RGB"`` or ``"L"`` (for grayscale).
+
+    ``palette`` is typically 256 or 256*3=768 bytes, for grayscale and RGB color
+    respectively, with each unit/triplet being the grayscale/RGB triplet values.
+    """
+
+    base_colorspace: str
+    palette: bytes
+
+
 class PdfImageBase(ABC):
 
     SIMPLE_COLORSPACES = {'/DeviceRGB', '/DeviceGray', '/CalRGB', '/CalGray'}
     MAIN_COLORSPACES = SIMPLE_COLORSPACES | {'/DeviceCMYK', '/CalCMYK', '/ICCBased'}
+    PRINT_COLORSPACES = {'/Separation', '/DeviceN'}
 
     @abstractmethod
     def _metadata(self, name, type_, default):
-        raise NotImplementedError()
+        ...
 
     @property
     def width(self):
@@ -131,7 +152,7 @@ class PdfImageBase(ABC):
         return self._metadata('DecodeParms', dict_or_array_dict, [])
 
     @property
-    def colorspace(self):
+    def colorspace(self) -> Optional[str]:
         """PDF name of the colorspace that best describes this image"""
         if self.image_mask:
             return None  # Undefined for image masks
@@ -142,8 +163,14 @@ class PdfImageBase(ABC):
                 subspace = self._colorspaces[1]
                 if isinstance(subspace, str) and subspace in self.MAIN_COLORSPACES:
                     return subspace
-                if isinstance(subspace, list) and subspace[0] == '/ICCBased':
+                if isinstance(subspace, list) and subspace[0] in (
+                    '/ICCBased',
+                    '/DeviceN',
+                ):
                     return subspace[0]
+            if self._colorspaces[0] == '/DeviceN':
+                return '/DeviceN'
+
         raise NotImplementedError(
             "not sure how to get colorspace: " + repr(self._colorspaces)
         )
@@ -158,18 +185,39 @@ class PdfImageBase(ABC):
     @property
     @abstractmethod
     def is_inline(self):
-        raise NotImplementedError()
+        ...
 
     @property
     @abstractmethod
     def icc(self):
-        raise NotImplementedError()
+        ...
 
     @property
     def indexed(self):
         """``True`` if the image has a defined color palette"""
         return '/Indexed' in self._colorspaces
 
+    def _colorspace_has_name(self, name):
+        try:
+            cs = self._colorspaces
+            if cs[0] == '/Indexed' and cs[1][0] == name:
+                return True
+            if cs[0] == name:
+                return True
+        except (IndexError, AttributeError, KeyError):
+            pass
+        return False
+
+    @property
+    def is_device_n(self):
+        """``True`` if image has a /DeviceN (complex printing) colorspace"""
+        return self._colorspace_has_name('/DeviceN')
+
+    @property
+    def is_separation(self):
+        """``True`` if image has a /DeviceN (complex printing) colorspace"""
+        return self._colorspace_has_name('/Separation')
+
     @property
     def size(self):
         """Size of image as (width, height)"""
@@ -199,7 +247,11 @@ class PdfImageBase(ABC):
         """
 
         m = ''
-        if self.indexed:
+        if self.is_device_n:
+            m = 'DeviceN'
+        elif self.is_separation:
+            m = 'Separation'
+        elif self.indexed:
             m = 'P'
         elif self.bits_per_component == 1:
             m = '1'
@@ -238,12 +290,8 @@ class PdfImageBase(ABC):
         return list(zip_longest(self.filters, self.decode_parms, fillvalue={}))
 
     @property
-    def palette(self):
-        """Retrieves the color palette for this image
-
-        Returns:
-            tuple (base_colorspace: str, palette: bytes)
-        """
+    def palette(self) -> Optional[PaletteData]:
+        """Retrieves the color palette for this image if applicable."""
 
         if not self.indexed:
             return None
@@ -251,53 +299,78 @@ class PdfImageBase(ABC):
             _idx, base, _hival, lookup = self._colorspaces
         except ValueError as e:
             raise ValueError('Not sure how to interpret this palette') from e
-        if self.icc:
+        if self.icc or self.is_device_n or self.is_separation:
             base = str(base[0])
         else:
             base = str(base)
         lookup = bytes(lookup)
-        if not base in self.MAIN_COLORSPACES:
-            raise NotImplementedError("not sure how to interpret this palette")
+        if base not in self.MAIN_COLORSPACES and base not in self.PRINT_COLORSPACES:
+            raise NotImplementedError(f"not sure how to interpret this palette: {base}")
         if base == '/DeviceRGB':
             base = 'RGB'
         elif base == '/DeviceGray':
             base = 'L'
         elif base == '/DeviceCMYK':
             base = 'CMYK'
+        elif base == '/DeviceN':
+            base = 'DeviceN'
+        elif base == '/Separation':
+            base = 'Separation'
         elif base == '/ICCBased':
             base = self._approx_mode_from_icc()
-        return base, lookup
+        return PaletteData(base, lookup)
 
     @abstractmethod
     def as_pil_image(self):
-        raise NotImplementedError()
+        ...
 
     @staticmethod
-    def _unstack_compression(buffer, filters):
-        """Remove stacked compression where it appears.
-
-        Stacked compression means when an image is set to:
-            ``[/FlateDecode /DCTDecode]``
-        for example.
-
-        Only Flate can be stripped off the front currently.
+    def _remove_simple_filters(obj, filters):
+        """Remove simple lossless compression where it appears.
 
         Args:
-            buffer (pikepdf._qpdf.Buffer): the compressed image data
+            obj (pikepdf.Stream): the compressed object
             filters (list of str): all files on the data
         """
-        data = memoryview(buffer)
-        while len(filters) > 1 and filters[0] == '/FlateDecode':
-            data = decompress(data)
-            filters = filters[1:]
-        return data, filters
+
+        COMPLEX_FILTERS = {
+            '/DCTDecode',
+            '/JPXDecode',
+            '/JBIG2Decode',
+            '/CCITTFaxDecode',
+        }
+
+        idx = [n for n, item in enumerate(filters) if item in COMPLEX_FILTERS]
+        if idx:
+            if len(idx) > 1:
+                raise NotImplementedError(
+                    f"Object {obj.objgen} has compound complex filters: {filters}. "
+                    "We cannot decompress this."
+                )
+            simple_filters = filters[: idx[0]]
+            complex_filters = filters[idx[0] :]
+        else:
+            simple_filters = filters
+            complex_filters = []
+
+        if not simple_filters:
+            return obj.read_raw_bytes(), complex_filters
+
+        original_filters = obj.Filter
+        try:
+            obj.Filter = Array([Name(s) for s in simple_filters])
+            data = obj.read_bytes(StreamDecodeLevel.specialized)
+        finally:
+            obj.Filter = original_filters
+
+        return data, complex_filters
 
 
 class PdfImage(PdfImageBase):
     """Support class to provide a consistent API for manipulating PDF images
 
-    The data structure for images inside PDFs is irregular and flexible,
-    making it difficult to work with without introducing errors for less
+    The data structure for images inside PDFs is irregular and complex,
+    making it difficult to use without introducing errors for less
     typical cases. This class addresses these difficulties by providing a
     regular, Pythonic API similar in spirit (and convertible to) the Python
     Pillow imaging library.
@@ -400,7 +473,7 @@ class PdfImage(PdfImageBase):
                     ) from e
         return self._icc
 
-    def _extract_direct(self, *, stream):
+    def _extract_direct(self, *, stream) -> str:
         """Attempt to extract the image directly to a usable image file
 
         If there is no way to extract the image without decompressing or
@@ -428,9 +501,7 @@ class PdfImage(PdfImageBase):
             ct = self.filter_decodeparms[0][1].get('/ColorTransform', DEFAULT_CT_CMYK)
             return self.mode == 'CMYK' and ct == DEFAULT_CT_CMYK
 
-        data, filters = self._unstack_compression(
-            self.obj.get_raw_stream_buffer(), self.filters
-        )
+        data, filters = self._remove_simple_filters(self.obj, self.filters)
 
         if filters == ['/CCITTFaxDecode']:
             if self.colorspace == '/ICCBased':
@@ -448,24 +519,56 @@ class PdfImage(PdfImageBase):
 
         raise NotExtractableError()
 
-    def _extract_transcoded(self):
+    def _extract_transcoded(self) -> Image.Image:
+        # Reminder Pillow palette byte order unintentionally changed in 8.3.0
+        # https://github.com/python-pillow/Pillow/issues/5595
+        # 8.2.0: all aligned by channel (very nonstandard)
+        # 8.3.0: all channels for one color followed by the next color (e.g. RGBRGBRGB)
+
         im = None
+        if self.mode in {'DeviceN', 'Separation'}:
+            raise HifiPrintImageNotTranscodableError()
+
         if self.mode == 'RGB' and self.bits_per_component == 8:
             # No point in accessing the buffer here, size qpdf decodes to 3-byte
             # RGB and Pillow needs RGBX for raw access
-            data = self.read_bytes()
-            im = Image.frombytes('RGB', self.size, data)
+            im = Image.frombuffer(
+                'RGB', self.size, self.read_bytes(), 'raw', 'RGB', 0, 1
+            )
+        elif self.mode == 'CMYK' and self.bits_per_component == 8:
+            im = Image.frombuffer(
+                'CMYK', self.size, self.get_stream_buffer(), 'raw', 'CMYK', 0, 1
+            )
         elif self.mode in ('L', 'P') and self.bits_per_component == 8:
             buffer = self.get_stream_buffer()
             stride = 0  # tell Pillow to calculate stride from line width
             ystep = 1  # image is top to bottom in memory
             im = Image.frombuffer('L', self.size, buffer, "raw", 'L', stride, ystep)
-            if self.mode == 'P':
+            if self.mode == 'P' and self.palette is not None:
                 base_mode, palette = self.palette
-                if base_mode in ('RGB', 'L'):
+                if base_mode == 'RGB':
                     im.putpalette(palette, rawmode=base_mode)
+                elif base_mode == 'L':
+                    # Pillow does not fully support palettes with rawmode='L'.
+                    # Convert to RGB palette.
+                    gray_palette = palette
+                    palette = b''
+                    for entry in gray_palette:
+                        palette += bytes([entry]) * 3
+                    im.putpalette(palette, rawmode='RGB')
+                elif base_mode == 'CMYK':
+                    # Pillow does not support CMYK with palettes; convert manually
+                    with memoryview(buffer) as mv:
+                        output = bytearray(4 * len(mv))
+                        for n, pal_idx in enumerate(mv):
+                            output[4 * n : 4 * (n + 1)] = palette[
+                                4 * pal_idx : 4 * (pal_idx + 1)
+                            ]
+                    im = Image.frombuffer(
+                        'CMYK', self.size, data=output, decoder_name='raw'
+                    )
                 else:
-                    raise NotImplementedError('palette with ' + base_mode)
+                    raise NotImplementedError(f'palette with {base_mode}')
         elif self.bits_per_component == 1:
             if self.filters and self.filters[0] == '/JBIG2Decode':
                 if not jbig2.jbig2dec_available():
@@ -479,30 +582,36 @@ class PdfImage(PdfImageBase):
                 data = self.read_bytes()
                 im = Image.frombytes('1', self.size, data)
         else:
-            raise UnsupportedImageTypeError(repr(self))
+            raise UnsupportedImageTypeError(repr(self) + ", " + repr(self.obj))
 
-        if self.mode == 'P' and self.bits_per_component == 1:
+        if (
+            self.mode == 'P'
+            and self.bits_per_component == 1
+            and self.palette is not None
+        ):
             # Fix paletted 1-bit images
             base_mode, palette = self.palette
             if base_mode == 'RGB' and palette != b'\x00\x00\x00\xff\xff\xff':
                 im = im.convert('P')
                 im.putpalette(palette, rawmode=base_mode)
                 gp = im.getpalette()
-                gp[765:768] = gp[3:6]  # work around Pillow bug
-                im.putpalette(gp)
+                if gp:
+                    gp[765:768] = gp[3:6]  # work around Pillow bug
+                    im.putpalette(gp)
             elif base_mode == 'L' and palette != b'\x00\xff':
                 im = im.convert('P')
                 im.putpalette(palette, rawmode=base_mode)
                 gp = im.getpalette()
-                gp[255] = gp[1]  # work around Pillow bug
-                im.putpalette(gp)
+                if gp:
+                    gp[255] = gp[1]  # work around Pillow bug
+                    im.putpalette(gp)
 
         if self.colorspace == '/ICCBased':
             im.info['icc_profile'] = self.icc.tobytes()
 
         return im
 
-    def _extract_to_stream(self, *, stream):
+    def _extract_to_stream(self, *, stream) -> str:
         """Attempt to extract the image to a stream
 
         If possible, the compressed data is extracted and inserted into
@@ -514,7 +623,7 @@ class PdfImage(PdfImageBase):
             stream: Writable stream to write data to
 
         Returns:
-            str: The file format extension
+            The file format extension.
         """
 
         try:
@@ -522,19 +631,26 @@ class PdfImage(PdfImageBase):
         except NotExtractableError:
             pass
 
+        im = None
         try:
             im = self._extract_transcoded()
-            if im:
+            if im.mode == 'CMYK':
+                im.save(stream, format='tiff', compression='tiff_adobe_deflate')
+                return '.tiff'
+            elif im:
                 im.save(stream, format='png')
                 return '.png'
         except PdfError as e:
-            if 'getStreamData called on unfilterable stream' in str(e):
+            if 'called on unfilterable stream' in str(e):
                 raise UnsupportedImageTypeError(repr(self)) from e
             raise
+        finally:
+            if im:
+                im.close()
 
         raise UnsupportedImageTypeError(repr(self))
 
-    def extract_to(self, *, stream=None, fileprefix=''):
+    def extract_to(self, *, stream=None, fileprefix='') -> str:
         """Attempt to extract the image directly to a usable image file
 
         If possible, the compressed data is extracted and inserted into
@@ -547,6 +663,8 @@ class PdfImage(PdfImageBase):
         When saving the image to a file, use a temporary filename, and then
         rename the file to its final name based on the returned file extension.
 
+        Images might be saved as any of .png, .jpg, or .tiff.
+
         Examples:
 
             >>> im.extract_to(stream=bytes_io)
@@ -564,9 +682,6 @@ class PdfImage(PdfImageBase):
             If *fileprefix* was provided, then the fileprefix with the
             appropriate extension. If no *fileprefix*, then an extension
             indicating the file type.
-
-        Return type:
-            str
         """
 
         if bool(stream) == bool(fileprefix):
@@ -590,11 +705,10 @@ class PdfImage(PdfImageBase):
         """Access this image with the buffer protocol"""
         return self.obj.get_stream_buffer(decode_level=decode_level)
 
-    def as_pil_image(self):
-        """Extract the image as a Pillow Image, using decompression as necessary
+    def as_pil_image(self) -> Image.Image:
+        """Extract the image as a Pillow Image, using decompression as necessary.
 
-        Returns:
-            PIL.Image.Image
+        Caller must close the image.
         """
         try:
             bio = BytesIO()
@@ -706,12 +820,14 @@ class PdfImage(PdfImageBase):
     def _repr_png_(self):
         """Display hook for IPython/Jupyter"""
         b = BytesIO()
-        im = self.as_pil_image()
-        im.save(b, 'PNG')
-        return b.getvalue()
+        with self.as_pil_image() as im:
+            im.save(b, 'PNG')
+            return b.getvalue()
 
 
 class PdfJpxImage(PdfImage):
+    """Support class for JPEG 2000 images. Implements the same API as :class:`PdfImage`."""
+
     def __init__(self, obj):
         super().__init__(obj)
         self._jpxpil = self.as_pil_image()
@@ -726,9 +842,7 @@ class PdfJpxImage(PdfImage):
         )
 
     def _extract_direct(self, *, stream):
-        data, filters = self._unstack_compression(
-            self.obj.get_raw_stream_buffer(), self.filters
-        )
+        data, filters = self._remove_simple_filters(self.obj, self.filters)
         if filters != ['/JPXDecode']:
             raise UnsupportedImageTypeError(self.filters)
         stream.write(data)
@@ -770,7 +884,7 @@ class PdfJpxImage(PdfImage):
 
 
 class PdfInlineImage(PdfImageBase):
-    """Support class for PDF inline images"""
+    """Support class for PDF inline images. Implements the same API as :class:`PdfImage`."""
 
     # Inline images can contain abbreviations that we write automatically
     ABBREVS = {
@@ -794,7 +908,7 @@ class PdfInlineImage(PdfImageBase):
     }
     REVERSE_ABBREVS = {v: k for k, v in ABBREVS.items()}
 
-    def __init__(self, *, image_data, image_object: tuple):
+    def __init__(self, *, image_data: Object, image_object: tuple):
         """
         Args:
             image_data: data stream for image, extracted from content stream
